

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>


#include <strings.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/sysctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <linux/un.h>
#include <stdint.h>

#include "babeld.h"
#include "interface.h"
#include "neighbour.h"
#include "util.h"
#include "re6st.h"


/*
 * Data received from the UPDATECOST message
 */
struct re6st_cost_modifier {
    unsigned char ipv6[16];
    uint32_t n_interface_index;   // network order
    uint16_t n_cost_multiplier;   // network order
    uint16_t padding;           // explicit alignment
};


static void
re6st_set_nonblock(int socket)
{
    /* Set socket to non-blocking */

    int flags = 0;

    if ((flags = fcntl(socket, F_GETFL, 0)) < 0)
    {
        perror("fcntl");
        exit(1);
    }

    if (fcntl(socket, F_SETFL, flags | O_NONBLOCK) < 0)
    {
        perror("fcntl");
        exit(1);
    }
}


static void
re6st_set_block(int socket)
{
    /* Set socket to blocking */

    int flags = 0;

    if ((flags = fcntl(re6st_socket, F_GETFL, 0)) < 0)
    {
        perror("fcntl");
        exit(1);
    }

    if (fcntl(re6st_socket, F_SETFL, flags & ~O_NONBLOCK) < 0)
    {
        perror("fcntl");
        exit(1);
    }
}



int
re6st_setup_socket(int setup)
{
    int re6st_len;
    const char *re6st_file = "/var/lib/babel-re6st.sock";
    struct sockaddr_un re6st_local_addr;

    if(setup) {
        debugf("setting up RE6ST socket\n");

        if(re6st_socket < 0) {
            re6st_socket = socket(AF_UNIX, SOCK_STREAM, 0);
            if(re6st_socket < 0) {
                perror("re6st_socket error");
                return -1;
            }
        }

        memset(&re6st_local_addr, 0, sizeof(struct sockaddr_un));

        re6st_local_addr.sun_family = AF_UNIX;
        strncpy(re6st_local_addr.sun_path, re6st_file, UNIX_PATH_MAX);
        unlink(re6st_local_addr.sun_path);
        re6st_len = strlen(re6st_local_addr.sun_path) + sizeof(re6st_local_addr.sun_family);

        bind(re6st_socket, (struct sockaddr *)&re6st_local_addr, re6st_len);
        if (listen(re6st_socket, 5) == -1) {
            perror("re6st_socket listen failed");
            exit(1);
        }

        return 1;
    } else {
        close(re6st_socket);
        re6st_socket = -1;
        return 1;
    }
}




int
re6st_send_file(FILE *fp, int socket)
{
    fseek(fp, 0L, SEEK_END);

    uint32_t content_size = ftell(fp);

    char *content = malloc(content_size + 1);

    rewind(fp);

    if (!fread(content, 1, content_size, fp)) {
        if (ferror(fp)) {
            perror("Error reading from temp file");
            free(content);
            return 0;
        }
    };

    re6st_set_block(socket);

    uint32_t content_size_sent = htonl(content_size);

    if (send(socket, &content_size_sent, 4, 0) == -1) {
        perror("send");
        debugf("errno: %i", errno);
    };

    if (send(socket, content, content_size, 0) == -1) {
        perror("send");
        debugf("errno: %i", errno);
    };

    free(content);
    return 1;
}





static int
re6st_send_dump_tables(int socket)
{
    int ret;
    FILE *fp = tmpfile();

    dump_tables(fp);
    ret = re6st_send_file(fp, socket);
    fclose(fp);
    return ret;
}




void
apply_cost_modifiers(struct re6st_cost_modifier *costs, int element_count) {
    debugf("Received %i cost modifiers\n", element_count);
    int i;
    struct neighbour *neigh;
    for (i=0; i<element_count; i++) {
        struct re6st_cost_modifier *cost = (costs+i);
        int interface_index = ntohl(cost->n_interface_index);
        int cost_multiplier = ntohs(cost->n_cost_multiplier);
        debugf("Neigh %d: %02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x, interface: %i, cost: %i\n",
                        i,
                        cost->ipv6[0],
                        cost->ipv6[1],
                        cost->ipv6[2],
                        cost->ipv6[3],
                        cost->ipv6[4],
                        cost->ipv6[5],
                        cost->ipv6[6],
                        cost->ipv6[7],
                        cost->ipv6[8],
                        cost->ipv6[9],
                        cost->ipv6[10],
                        cost->ipv6[11],
                        cost->ipv6[12],
                        cost->ipv6[13],
                        cost->ipv6[14],
                        cost->ipv6[15],
                        interface_index,
                        cost_multiplier);
        FOR_ALL_NEIGHBOURS(neigh) {
            if (!memcmp(cost->ipv6, neigh->address, 16) && (interface_index == neigh->ifp->ifindex)) {
                debugf("Neighbour address and interface match. Setting cost multiplier to %i\n", cost_multiplier);
                neigh->cost_multiplier = cost_multiplier;
            }
        }
    }
}





int
re6st_accept_connection(int socket)
{
    re6st_set_nonblock(socket);

    debugf("Accept re6st connection on socket %d\n", socket);
    int newsock = -1;
    newsock = accept(socket, NULL, NULL);

    if (newsock < 0) {
        if (errno != EINTR && errno != EAGAIN) {
            perror("accept(re6st_socket)");
            return -1;
        }
        return 0;
    }

    int n = 0;
    uint32_t payload_size = 0;

    n = recv(newsock, &payload_size, 4, 0);
    payload_size = ntohl(payload_size);

    if (n == 0) {
        // remote has closed connection
        return 0;
    }

    debugf("PAYLOAD SIZE: %i\n", payload_size);

    char *payload;
    payload = malloc(payload_size + 1);
    int remaining = payload_size;

    while (remaining) {
        debugf("looping recv\n");
        n = recv(newsock, payload + payload_size - remaining, remaining, 0);
        if (n == 0) {
            // remote has closed connection
            free(payload);
            return 0;
        }
        if (n == -1) {
            if (errno == EAGAIN) {
                continue;
            } else {
                return 0;
            }
        }
        remaining -= n;
    };

    payload[payload_size] = 0;

    if (!strncmp(payload, "DUMP TABLES", payload_size)) {
        re6st_send_dump_tables(newsock);
    } else if (!strncmp(payload, "DUMP NEIGHBOURS", payload_size)) {
        dump_neighbours(newsock);
    } else if (!strncmp(payload, "UPDATE COSTS", 12)) {
        apply_cost_modifiers((struct re6st_cost_modifier *)(payload+12), (payload_size-12)/sizeof(struct re6st_cost_modifier));
    } else {
        debugf("Unknown protocol command %s", payload);
    }

    free(payload);

    close(newsock);

    return 1;
}
